
/*! 
    \file vlist.h
    \brief A documented header file containing the vlist 
*/


#ifndef VLIST_H
#define VLIST_H


#include <stdlib.h>
 
#ifdef _WIN32_WCE
#define VLIST_CDECL __cdecl
#else
#define VLIST_CDECL 
#endif
 
#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#ifndef DWFTK_STATIC

#ifdef  _DWFCORE_WIN32_SYSTEM
#ifndef DWFTK_EXPORT_API
#define _W3DTK_API __declspec(dllimport)
#else
#define _W3DTK_API __declspec(dllexport)
#endif
#else
#define _W3DTK_API
#endif

#else
#define _W3DTK_API
#endif

/*!
	The vlist is single linked list that stores void pointers.
	It has a cursor to traverse the list. 
*/
typedef struct vlist_node_s{
	void * item;
	struct vlist_node_s * next;
} vlist_node_t;

typedef struct vlist_s{
	struct vlist_node_s * head;
	struct vlist_node_s * tail;
	struct vlist_node_s * cursor;
	struct vlist_node_s * cursor_backlink;
	unsigned int cursor_index;
	unsigned int count;
	void *(VLIST_CDECL *malloc) (size_t);
	void (VLIST_CDECL *free) (void*);
} vlist_t;


/*!
    \brief This function creates and initalises a vlist structure.
	\param VHASH_CDECL* vlist_malloc A malloc function for vlist to use	
	\param VHASH_CDECL *vlist_free A free function for vlist to use	
	\return The vlist structure 
*/
_W3DTK_API vlist_t* new_vlist(
	void *  (VLIST_CDECL *  vlist_malloc) (size_t),
	void (VLIST_CDECL *vlist_free) (void *));

/*!
    \brief This function destroys and cleans up a vlist structure.
	\param vlist The list to delete
*/
_W3DTK_API void delete_vlist( 
	vlist_t* vlist); 

/*!
    \brief This function adds an item to the head of a vlist structure.
	\param vlist The list operate on
	\param item A void pointer to store 
*/
_W3DTK_API void vlist_add_first( 
	vlist_t* vlist, 
	void* item);

/*!
    \brief This function adds an item to the tail of a vlist structure.
	\param vlist The list operate on
	\param item A void pointer to store 
*/
_W3DTK_API void vlist_add_last( 
	vlist_t* vlist, 
	void* item);


/*!
    \brief This function adds an item before the cursor of a vlist structure.
	\param vlist The list operate on
	\param item A void pointer to store 
*/
_W3DTK_API void vlist_add_before_cursor( 
	vlist_t* vlist, 
	void* item);

/*!
    \brief This function adds an item after the cursor to the vlist structure.
	\param vlist The list operate on
	\param item A void pointer to store 
*/
_W3DTK_API void vlist_add_after_cursor(
	vlist_t* vlist,
	void *item);

/*!
    \brief This function removes an item from the vlist structure.
	\param vlist The list operate on
	\param item A void pointer to remove from the list 
	\return 0 if item not found ~0 if found
*/
_W3DTK_API int vlist_remove(
	vlist_t* vlist, 
	void* item);

/*!
    \brief This function removes the first item from the vlist structure.
	\param vlist The list operate on
	\return 0 if vlist is empty or first item pointer
*/
_W3DTK_API void * vlist_remove_first(
	vlist_t* vlist);

/*!
    \brief This function removes the next item after the cursor from the vlist structure.
	\param vlist The list operate on
	\return 0 if no such item exists else the item
*/
_W3DTK_API void * vlist_remove_cursor_next(
	vlist_t* vlist);

/*!
    \brief This function removes the item at the cursor from the vlist structure.
	\param vlist The list operate on
	\return 0 if no such item exists else the item
*/
_W3DTK_API void * vlist_remove_at_cursor(
	vlist_t* vlist);

/*!
    \brief This function moves the cursor to the front of the vlist structure.
	\param vlist The list operate on
*/
_W3DTK_API void vlist_reset_cursor(
	vlist_t* vlist);

/*!
    \brief This function returns item at the cursor but doesn't remove it.
	\param vlist The list operate on
	\return 0 if no such item exists else the item
*/
_W3DTK_API void * vlist_peek_cursor(
	vlist_t* vlist);

/*!
    \brief This function returns next item after the cursor but doesn't remove it.
	\param vlist The list operate on
	\return 0 if no such item exists else the item
*/
_W3DTK_API void * vlist_peek_cursor_next(
	vlist_t* vlist);

/*!
    \brief This function returns next item's next after the cursor but doesn't remove it.
	\param vlist The list operate on
	\return 0 if no such item exists else the item
*/
_W3DTK_API void * vlist_peek_cursor_next_next(
	vlist_t* vlist);

/*!
    \brief This function moves the cursor to the next postition in the vlist structure.
	\param vlist The list operate on
*/
_W3DTK_API void vlist_advance_cursor(
	vlist_t* vlist);

/*!
    \brief This function returns first item on the vlist but doesn't remove it.
	\param vlist The list operate on
	\return 0 if no such item exists else the item
*/
_W3DTK_API void * vlist_peek_first(
	vlist_t* vlist);

/*!
    \brief This function returns last item on the vlist but doesn't remove it.
	\param vlist The list operate on
	\return 0 if no such item exists else the item
*/
_W3DTK_API void * vlist_peek_last(
	vlist_t* vlist);

/*!
    \brief This function returns number of items in the vlist.
	\param vlist The list operate on
	\return Number of items in the list.
*/
_W3DTK_API unsigned long vlist_count(
	vlist_t* vlist);

/*!
    \brief This function checks the existence of a item in the vlist.
	\param vlist The list operate on
	\param item A void pointer look for
	\return 0 if item not found ~0 if found
*/
_W3DTK_API int vlist_item_exists(
	vlist_t* vlist, 
	void* item);

/*!
    \brief This function writes the items in the vlist into a array doesn't change the list .
	\param vlist The list operate on
	\param items An array of void pointers the size of number of items in the list
*/
_W3DTK_API void vlist_items_to_array(
	vlist_t* vlist, 
	void** items);

/*!
    \brief This function calls a function once for each item in the vlist. 
	\param vlist The list operate on
	\param function A pointer to a function taking an item and user data.
	\param user_data A pointer passed with each function call. 
*/
_W3DTK_API void vlist_map_function(
	vlist_t* vlist, 
	void(*function)(void*, void*),
	void * user_data);

/*!
    \brief This function peeks at the nth item in the vlist. 
	\param vlist The list operate on
	\param index A 0 based index into the vlist
	\return 0 if no such item exists else the item
*/
_W3DTK_API void* vlist_nth_item(
	vlist_t* vlist,
	unsigned long index);

/*!
    \brief This function reverses the items in the vlist.
	\param vlist The list operate on
*/
_W3DTK_API void vlist_reverse(
	vlist_t* vlist);

/*!
    \brief This function wipes out all items in the list
	\param vlist The list operate on
*/
_W3DTK_API void vlist_flush(
	vlist_t* vlist);

#ifdef __cplusplus
} /* extern "C" */
#endif /* __cplusplus */

#ifndef V_POINTER_SIZED_INT
#ifdef _WIN64
#define  V_POINTER_SIZED_INT    unsigned __int64
#else
#define  V_POINTER_SIZED_INT    unsigned long
#endif
#endif

/* these are used for the internal functions */
#define VLIST_RESET_CURSOR(XL) \
	(XL)->cursor = (XL)->head; \
	(XL)->cursor_index=0; \
	(XL)->cursor_backlink=0;

#define VLIST_PEEK_CURSOR(XL) \
    ((!(XL)->cursor)?0:(XL)->cursor->item)

#define	VLIST_ADVANCE_CURSOR(XL) \
    if((XL)->cursor) {\
		(XL)->cursor_backlink=(XL)->cursor; \
		(XL)->cursor = (XL)->cursor->next; \
		(XL)->cursor_index++; \
    }

#define LIST_REMOVE_AND_CONTINUE_ITERATION(XL) \
	vlist_remove_at_cursor((XL)); \
	continue;

#define START_LIST_ITERATION(TYPE, LIST) VLIST_RESET_CURSOR(LIST);\
				while (1) {\
				TYPE *temp;\
				if ((temp = (TYPE *)VLIST_PEEK_CURSOR(LIST)) == NULL)\
					break;

#define END_LIST_ITERATION(LIST)  VLIST_ADVANCE_CURSOR(LIST); }


/* template wrapper */
#ifdef __cplusplus
extern "C++" {

/*the following macro generates a compile-time errors
  if anyone tries to instantiate a VList with incompatible types */
#define VLIST_COMPILE_TIME_ASSERT(expr) { char unnamed[(expr)?1:0]; (void)unnamed; }

template <class T> class VList {
public:

	VList(struct vlist_s * vl) { m_vlist = vl; m_free_items=false; size_assert(); };

	VList() { m_vlist = new_vlist(malloc,free); m_free_items=false; size_assert(); };
	
	VList(void * (VLIST_CDECL *  vlist_malloc) (size_t), void (VLIST_CDECL *vlist_free) (void *)) 
		{m_vlist = new_vlist( vlist_malloc,vlist_free); m_free_items=false; size_assert(); };

	~VList() {
		if(m_free_items) {
			T tmp;
			while((tmp=RemoveFirst()))
				m_vlist->free((void*)(V_POINTER_SIZED_INT)tmp);
		}
		delete_vlist(m_vlist);
	};

	void AddFirst(T item) { vlist_add_first(m_vlist, (void *)(V_POINTER_SIZED_INT)item); };

	void AddLast(T item) { vlist_add_last(m_vlist, (void *)(V_POINTER_SIZED_INT)item); };

	void AddAfterCursor(T item) { vlist_add_after_cursor(m_vlist, (void *)(V_POINTER_SIZED_INT)item); };

	int Remove(T item) { return (vlist_remove(m_vlist, (void *)(V_POINTER_SIZED_INT)item)); };

	T RemoveFirst() { return ((T)(V_POINTER_SIZED_INT)vlist_remove_first(m_vlist)); };

 	void ResetCursor() { vlist_reset_cursor(m_vlist); };

 	T PeekCursor() { return ((T)(V_POINTER_SIZED_INT)vlist_peek_cursor(m_vlist)); };

 	T PeekCursorNext() { return ((T)(V_POINTER_SIZED_INT)vlist_peek_cursor_next(m_vlist)); };

	T PeekCursorNextNext() { return ((T)(V_POINTER_SIZED_INT)vlist_peek_cursor_next_next(m_vlist)); };

	void AdvanceCursor() { vlist_advance_cursor(m_vlist); };

	T PeekFirst() { return ((T)(V_POINTER_SIZED_INT)vlist_peek_first(m_vlist)); };

 	T PeekLast() { return ((T)(V_POINTER_SIZED_INT)vlist_peek_last(m_vlist)); };

 	unsigned long Count() { return (vlist_count(m_vlist)); };

	int ItemExists(T item) { return (vlist_item_exists(m_vlist, (void *)(V_POINTER_SIZED_INT)item)); };

	void ItemsToArray(T* items) {
		if (sizeof(T) != sizeof(V_POINTER_SIZED_INT)) {
			V_POINTER_SIZED_INT *tmp=new V_POINTER_SIZED_INT[m_vlist->count];
			vlist_items_to_array(m_vlist, (void **)tmp);
			for(unsigned int i=0; i<m_vlist->count; i++)
				items[i] = (T)tmp[i];
			delete[] tmp;
		}
		else
			vlist_items_to_array(m_vlist, (void **)items);
	};

	void MapFunction(void(*function)(T, void*), void const * const user_data) {
		unsigned long i = Count();
		ResetCursor();
		while(i){
			(*function)(PeekCursor(),(void*)user_data);
			AdvanceCursor();
			i--;
		}
	};

	T NthItem(unsigned long index) { return ((T)(V_POINTER_SIZED_INT)vlist_nth_item(m_vlist, index)); };
	
	void Reverse() { vlist_reverse(m_vlist); };
	
	void Flush() { vlist_flush(m_vlist); };

	void SetFreeItemsOnDelete() {m_free_items = true;};

private:
	struct vlist_s *m_vlist;
	bool m_free_items;

	/* VList cannot hold any type that cannot be converted to and from a (void *) */
	void size_assert() { VLIST_COMPILE_TIME_ASSERT(sizeof(T) <= sizeof(void *)); };

};
} /* extern "C++" */
#endif

#endif /*VLIST_H*/



